/** 
 * \file exceptionobject.hpp
 *  
 * \brief This  file contains the  definition of a class  for handling
 * exceptions.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef EXCEPTIONOBJECT_HPP
#define EXCEPTIONOBJECT_HPP

#include <iostream>     // std::cout, std::endl
#include <string>       // std::string
#include <stdexcept>    // std::exception


/**
 * \fn treat_exception( e )
 *
 * \brief A macro to write information about an exception.
 *
 * \param e An exception.
 */
#define treat_exception( e ) \
  std::cout << "Exception: " << e.get_description() << std::endl \
            << "File: " << e.get_file() << std::endl \
            << "Line: " << e.get_line() << std::endl



/** 
 * \defgroup CommonNameSpace Namespace common.
 * @{
 */

/**
 * \namespace common
 *
 * \brief The  namespace common contains the definition  of macros and
 * several  utility  classes, such  as  a  class  for representing  3D
 * vectors and  a class  for dealing with  exceptions in  a customized
 * way.
 */

namespace common {


  /**
   * \class ExceptionObject
   *
   * \brief This class extends the class exception of STL, providing a
   * customized way of dealing with exceptions.
   */
  class ExceptionObject : public std::exception {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn ExceptionObject( const char *file , unsigned int line , const char *desc )
     *
     * \brief Creates an instance of this class.
     *
     * \param file A file with the code that threw the exception.
     * \param line  The number  of line (in  the file) that  threw the
     * exception.
     * \param desc A description of the exception.
     */
    ExceptionObject(
		    const char *file = "Unknown" ,
		    unsigned int line = 0 ,
		    const char *desc = "None"
		   )
    {
      _file = file ;
      _desc = desc ;
      _line = line ;
    }


    /** 
     * \fn ExceptionObject( const std::string& file , unsigned int line , const std::string& desc )
     *
     * \brief Creates an instance of this class.
     *
     * \param file A file with the code that threw the exception.
     * \param line  The number  of line (in  the file) that  threw the
     * exception.
     * \param desc A description of the exception.
     */
    ExceptionObject(
		    const std::string& file ,
		    unsigned int line ,
		    const std::string& desc = "None"
		   )
    {
      _file = file ;
      _desc = desc ;
      _line = line ;
    }


    /** 
     * \fn ExceptionObject( const ExceptionObject &xpt )
     *
     * \brief Creates an instance  of this class from another instance
     * of this class.
     *
     * \param xpt Another instance of this class.
     */
    ExceptionObject( const ExceptionObject &xpt ) : exception()
    {
      _file = xpt._file ;
      _desc = xpt._desc ;
      _line = xpt._line ;
    }


    /** 
     * \fn ~ExceptionObject() 
     *
     * \brief Destroys an instance of this class.
     */
    virtual ~ExceptionObject() throw()
    {
    }
    

    /** 
     * \fn ExceptionObject& operator=( const ExceptionObject &xpt )
     *
     * \brief Copies the state of a given exception to this exception.
     *
     * \param xpt A given exception.
     */
    ExceptionObject& operator=( const ExceptionObject &xpt )
    {
      _desc = xpt._desc ;
      _file = xpt._file ;
      _line = xpt._line ;
      
      return *this ;
    }


    /** 
     * \fn virtual const char* get_name_of_class() const 
     *
     * \brief Gets the name of this class.
     *
     * \return The name of this class.
     */
    virtual const char* get_name_of_class() const 
    {
      return "ExceptionObject" ;
    }


    /** 
     * \fn virtual void set_description( const std::string& s ) 
     *
     * \brief Assigns a description to this exception.
     *
     * \param s A string containing an exception description.
     */
    virtual void set_description( const std::string& s ) 
    { 
      _desc = s ; 
    }
    

    /** 
     * \fn virtual void set_description( const char* s ) 
     *
     * \brief Assigns a description to this exception.
     *
     * \param s A string containing an exception description.
     */
    virtual void set_description( const char *s )       
    { 
      _desc = s ; 
    }


    /** 
     * \fn virtual const char* get_description() const 
     *
     * \brief Gets the description of this exception.
     *
     * \return The description of this exception.
     */
    virtual const char* get_description() const 
    { 
      return _desc.c_str() ;
    }


    /** 
     * \fn virtual const char* get_file() const
     *
     * \brief Gets the name of the file containing the code that threw
     * the exception.
     *
     * \return The name of the file containing the code that threw the
     * exception.
     */
    virtual const char* get_file() const 
    { 
      return _file.c_str() ;
    }


    /** 
     * \fn virtual unsigned int get_line() const 
     *
     * \brief Gets the number of the line that threw the exception.
     *
     * \return The number of the line that threw the exception.
     */
    virtual unsigned int get_line() const 
    {
      return _line ;
    }
    

    /** 
     * \fn virtual const char* what() const throw()
     *
     * \brief Gets a description for the exception.
     *
     * \return A description for the exception.
     */
    virtual const char* what() const throw()
    { 
      return _desc.c_str() ;
    }
    
    
  private:
    // -----------------------------------------------------------------
    //
    // Private attributes
    //
    // -----------------------------------------------------------------
    
    std::string  _desc ;    ///< Description of the error
    std::string  _file ;    ///< File where the error occured
    unsigned int _line ;    ///< Line of the file where the error occured
    
  } ;


}         // end of namespace common


/** @} */ //end of group class.


#endif   // EXCEPTIONOBJECT_HPP
  
