/** 
 * \file mathmacros.hpp
 *  
 * \brief This file contains the definition of several math macros.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef MATHMACROS_HPP
#define MATHMACROS_HPP

#include <cmath>  // fabs(), acos()


#define PI	         acos( -1 )
#define D_HUGE		 1e40
#define SQR( x )	 ( ( x ) * ( x ) )
#define CUBE( x )	 ( ( x ) * ( x ) * ( x ) )
#define INVERSE( x )     ( 1. / ( x ) )
#define MID( x )	 ( ( x ) / 2. )
#define DUP( x )	 ( ( x ) * 2. )
#define ABS( x )	 ( fabs( x ) )
#define MOD( x )	 ( ABS( x ) )
#define ISZERO( x )	 ( ABS( x ) <= 1.e-20 )
#define ISNEGATIVE( x )	 ( ( x ) < 0. )
#define ISPOSITIVE( x )	 ( ( x ) > 0. )
#define ISEQUAL( x , y ) ( ISZERO( ( x ) - ( y ) ) )
#define IS2PI( x )	 ( ISEQUAL( ( x ) , 2 * PI ) )
#define MIN( x , y )	 ( ( x ) < ( y ) ? ( x ) : ( y ) )
#define MAX( x , y )	 ( ( x ) > ( y ) ? ( x ) : ( y ) )
#define TORADIANS( x )	 ( ( x ) * PI / 180 )
#define TODEGREES( x )	 ( ( x ) * 180 / PI )
#define SIGN( x )	 ( ISNEGATIVE( x ) ? -1 : 1 )


/** 
 * \defgroup CommonNameSpace Namespace common.
 * @{
 */

/**
 * \namespace common
 *
 * \brief The  namespace common contains the definition  of macros and
 * several  utility  classes, such  as  a  class  for representing  3D
 * vectors and  a class  for dealing with  exceptions in  a customized
 * way.
 */

namespace common {

  //
  // Compare functions templates
  //
  template <class T> inline bool
  is_equal( T a , T b ) {
    return a == b ;
  }
  
  template <class T> inline bool
  is_zero( T a ) {
    return a == 0 ;
  }
  
  template <class T> inline bool
  is_positive( T a ) {
    return a > 0 ;
  }
  
  template <class T> inline bool
  is_negative( T a ) {
    return a < 0 ;
  }
  
  //
  // Override compare functions templates for double type
  //
  inline bool
  is_equal( double a , double b ) {
    return ISEQUAL( a , b ) ;
  }

  inline bool
  is_zero( double a ) {
    return ISZERO( a ) ;
  }
  
  inline bool
  is_positive( double a ) {
    return a > 0 && !is_zero( a ) ;
  }
  
  inline bool
  is_negative( double a ) {
    return a < 0 && !is_zero( a ) ;
  }

  //
  // Other compare functions for double type
  //
  inline bool
  is_lesser( double a , double b ) {
    return a < b && !is_equal( a , b ) ;
  }

  inline bool
  is_lesser_equal( double a , double b ) {
    return a < b || is_equal( a , b ) ;
  }

  inline bool
  is_greater( double a , double b ) {
    return a > b && !is_equal( a , b ) ;
  }

  inline bool
  is_greater_equal( double a , double b ) {
    return a > b || is_equal( a , b ) ;
  }

  //
  // Other functions
  //
  inline double
  to_radians( double a ) {
    return TORADIANS( a ) ;
  }

  inline double
  to_degrees( double a ) {
    return TODEGREES( a ) ;
  }

  inline double
  sqr( double a ) {
    return a * a ;
  }

  inline double
  mysqrt( double a ) {
    if ( a < 0 ) return 0 ;
    return sqrt( a ) ;
  }

}          // end of namespace common


/** @} */ //end of group class.


#endif	 // MATHMACROS_HPP
