/** 
 * \file svddcmp.cpp
 *  
 * \brief This file contains the definition of a class that contains a
 * method for  computing the Singular  Value Decomposition (SVD)  of a
 * real matrix.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "svddcmp.hpp"          // SVDdcmp

#include "exceptionobject.hpp"  // commonlib::ExceptionObject
#include "mathmacros.hpp"       // SQR, MAX, MIN

#include <cassert>              // assert()
#include <sstream>              // std::stringstream 
#include <cmath>                // fabs(), sqrt()


/** 
 * \defgroup NumericsNameSpace Namespace numerics
 * @{
 */

/**
 * \namespace numerics
 *
 * \brief  The  namespace  numerics  contains  a set  of  classes  and
 * functions for numerical computing.
 */

namespace numerics {

  using common::ExceptionObject ;


  /** 
   * \fn SVDdcmp::SVDdcmp( double** a , int m , int n ) 
   *
   * \brief Creates an instance of this class.
   *
   * \param a A pointer to a real matrix.
   * \param m The number of rows of the given matrix.
   * \param n The number of columns of the given matrix.
   */
  SVDdcmp::SVDdcmp(
		   double** a ,
		   int m ,
		   int n
		  )
  {
    //
    // Initialize dimensions of matrix _A .
    //
    assert( ( m > 0 ) && ( n > 0 ) ) ;

    _m = m ;
    _n = n ;

    //
    // Allocate memory for matrices _A and _V, and vector _W.
    //
    allocate() ;

    //
    // Copy data from "a" to _A, and fill in _W with zeros.
    //
    for ( int i = 0 ; i < _m ; i++ ) {
      for ( int j = 0 ; j < _n ; j++ ) {
        _A[ i ][ j ] = a[ i ][ j ] ;
      }
    }

    //
    // Fill in matrix _V and vector _W with zeros.
    //
    for ( int i = 0 ; i < _n ; i++ ) {
      _W[ i ] = 0 ;
      for ( int j = 0 ; j < _n ; j++ ) {
        _V[ i ][ j ] = 0 ;
      }
    }

    //
    // Compute the SVD decomposition of matrix _A.
    //
    decomp() ;

    //
    // Find the largest singular value of _A.
    //
    double wmax = 0 ;
    for ( int i = 0 ; i < _n ; i++ ) {
      if ( _W[ i ] > wmax ) {
	wmax = _W[ i ] ;
      }
    }

    //
    // Compute the threshold for the nonzero singular values.
    //
    double wmin = wmax * 1.0e-12 ;
    for ( int i = 0 ; i < _n ; i++ ) {
      if ( _W[ i ] < wmin ) {
	_W[ i ] = 0 ;
      }
    }

    return ;
  }


  /** 
   * \fn SVDdcmp::SVDdcmp( const SVDdcmp& s ) 
   *
   * \brief  Creates an  instance  of this  class  by cloning  another
   * instance of this class.
   *
   * \param s A given instance of this class
   */
  SVDdcmp::SVDdcmp( const SVDdcmp& s )
  {
    //
    // Initialize dimensions of matrix _A.
    //
    _m = s._m ;
    _n = s._n ;

    //
    // Allocate memory for matrices _A and _V, and vector _W.
    //
    allocate() ;

    //
    // Copy values for _A, _V, and _W.
    //
    for ( int i = 0 ; i < _m ; i++ ) {
      for ( int j = 0 ; j < _n ; j++ ) {
        _A[ i ][ j ] = s._A[ i ][ j ] ;
      }
    }

    for ( int i = 0 ; i < _n ; i++ ) {
      _W[ i ] = s._W[ i ] ;
      for ( int j = 0 ; j < _n ; j++ ) {
        _V[ i ][ j ] = s._V[ i ][ j ] ;
      }
    }

    return ;
  }


  /** 
   * \fn SVDdcmp::~SVDdcmp() 
   *
   * \brief Destroys an instance of this class.
   */
  SVDdcmp::~SVDdcmp()
  {
    //
    // Release memory allocated for matrices _A and _V, and vector _W.
    //
    deallocate() ;

    return ;
  }


  /** 
   * \fn double SVDdcmp::matU( int i , int j ) const throw ( ExceptionObject )
   *
   * \brief Returns the element of the i-th row and j-th column of the
   * orthogonal matrix \f$U\f$ of  the SVD decomposition, \f$ U \Sigma
   * S V^t \f$, of the given real matrix, \f$A\f$.
   *
   * \param i The row index of an element.
   * \param j The column index of an element.
   */
  double SVDdcmp::matU( 
		       int i , 
		       int j 
		      ) 
    const throw ( ExceptionObject )
  {
    //
    // To save memory, matrix U is stored in _A.
    //
    return matA( i , j ) ;
  }


  /** 
   * \fn double SVDdcmp::matV( int i , int j ) const throw ( ExceptionObject )
   *
   * \brief Returns the element of the i-th row and j-th column of the
   * orthogonal matrix \f$V\f$ of  the SVD decomposition, \f$ U \Sigma
   * S V^t \f$, of the given real matrix, \f$A\f$.
   *
   * \param i The row index of an element.
   * \param j The column index of an element.
   */
  double
  SVDdcmp::matV( 
		int i , 
		int j 
	       ) 
    const throw ( ExceptionObject )
  {
    if ( ( i < 0 ) || ( i >= _n ) || ( j < 0 ) || ( j >= _n ) ) {
      std::stringstream ss(
			   std::stringstream::in | 
			   std::stringstream::out
			  ) ;
      ss << "Bad element index" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }

    return _V[ i ][ j ] ;
  }


  /** 
   * \fn double SVDdcmp::vecW( int i ) const throw ( ExceptionObject )
   *
   * \brief Returns the  i-th singular value of the  given real matrix
   * for which the SVD decomposition has been computed.
   *
   * \param i The index of a singular value.
   */
  double
  SVDdcmp::vecW( int i ) const throw ( ExceptionObject )
  {
    if ( ( i < 0 ) || i >= _n ) {
      std::stringstream ss(
			   std::stringstream::in | 
			   std::stringstream::out
			  ) ;
      ss << "Bad element index" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }

    return _W[ i ] ;
  }


  /** 
   * \fn void SVDdcmp::lsq_solve( double* b , double* x ) const
   *
   * \brief Solves the linear least  squares problem \f$\| A \cdot X -
   * B \| \f$,  where \f$ A \f$  is the real matrix for  which the SVD
   * decomposition has  been computed, and \f$  B \f$ is  a given real
   * vector.
   *
   * \param b A pointer to the column vector \f$ B \f$.
   * \param x A pointer to the solution vector  \f$ X \f$.
   */
  void
  SVDdcmp::lsq_solve( 
		     double* b , 
		     double* x 
		    ) 
    const
  {
    //
    // Compute \f$ A X = B \f$ by backsubstitution.
    //
    bksb( b , x ) ;

    return ;
  }


  // -----------------------------------------------------------------
  //
  // Private methods
  //
  // -----------------------------------------------------------------
  

  /** 
   * \fn double SVDdcmp::matA( int i , int j ) const throw ( ExceptionObject )
   *
   * \brief Returns the element of the i-th row and j-th column of the
   * real  matrix \f$A\f$  for  which the  SVD  decomposition will  be
   * computed.
   *
   * \param i The row index of an element.
   * \param j The column index of an element.
   */
  double SVDdcmp::matA( 
		       int i , 
		       int j 
		      ) 
    const throw ( ExceptionObject )
  {
    if ( ( i < 0 ) || ( j < 0 ) || ( i >= _m ) || ( j >= _n ) ) {
      std::stringstream ss(
			   std::stringstream::in | 
			   std::stringstream::out
			  ) ;
      ss << "Bad element index" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }

    return _A[ i ][ j ] ;
  }


  /** 
   * \fn void SVDdcmp::allocate()
   *
   * \brief Allocates memory for the matrices \f$ A \f$ and \f$ V \f$,
   * and the  vector \f$ W  \f$, where \f$  V\f$ and \f$ W\f$  are the
   * orthogonal   matrix   and  the   singular   value   of  the   SVD
   * decomposition, \f$ U \Sigma S  V^t \f$, of the given real matrix,
   * \f$A\f$.
   */
  void
  SVDdcmp::allocate()
  {
    _A = new double*[ _m ] ;
    _V = new double*[ _n ] ;
    _W = new double [ _n ] ;

    for ( int i = 0 ; i < _m ; ++i ) {
      _A[ i ] = new double[ _n ] ;
    }

    for ( int i = 0 ; i < _n ; ++i ) {
      _V[ i ] = new double[ _n ] ;
    }

    return ;
  }


  /** 
   * \fn void SVDdcmp::deallocate()
   *
   * \brief Deallocates  the memory allocated  for the matrices  \f$ A
   * \f$ and \f$ V  \f$, and the vector \f$ W \f$,  where \f$ V\f$ and
   * \f$ W\f$ are the orthogonal  matrix and the singular value of the
   * SVD  decomposition, \f$ U  \Sigma S  V^t \f$,  of the  given real
   * matrix, \f$A\f$.
   */
  void
  SVDdcmp::deallocate()
  {
    if ( _A != 0 ) {
      for ( int i = 0 ; i < _m ; ++i ) {
        if ( _A[ i ] != 0 ) {
	  delete[] _A[ i ] ;
	}
      }

      delete[] _A ;
    }

    if ( _V != 0 ) {
      for ( int i = 0 ; i < _n ; ++i ) {
        if ( _V[ i ] != 0 ) {
	  delete[] _V[ i ] ;
	}
      }

      delete[] _V ;
    }

    if ( _W ) {
      delete[] _W ;
    }

    return ;
  }


  /** 
   * \fn void SVDdcmp::decomp() const throw ( ExceptionObject )
   *
   * \brief  Computes  the  SVD  decomposition of  the  given,  real
   * matrix.
   */
  void
  SVDdcmp::decomp() const throw ( ExceptionObject )
  {
    bool flag ;
    int i , j , k , l = 0 , nm = 0 ;
    double anorm , c , f , g , h , s , scale , x , y , z ;

    double* rv1 = new double[ _n ] ;

    g = scale = anorm = 0.0 ;

    for ( i = 0 ; i < _n ; i++ ) {
      l = i + 2 ;
      rv1[ i ] = scale * g ;
      g = s = scale = 0.0 ;
      if ( i < _m ) {
        for ( k = i ; k < _m ; k++ ) {
          scale += fabs( _A[ k ][ i ] ) ;
        }
	if ( scale != 0.0 ) {
          for ( k = i ; k < _m ; k++ ) {
	    _A[ k ][ i ] /= scale ;
            s += _A[ k ][ i ] * _A[ k ][ i ] ;
          }

          f = _A[ i ][ i ] ;
          g = -sign( sqrt( s ) , f ) ;
          h = f * g - s ;
          _A[ i ][ i ] = f - g ;
	  
          for ( j = l - 1 ; j < _n ; j++ ) {
            for ( s = 0.0 , k = i ; k < _m ; k++ ) {
              s += _A[ k ][ i ] * _A[ k ][ j ] ;
            }

            f = s / h ;

            for ( k = i ; k < _m ; k++ ) {
              _A[ k ][ j ] += f * _A[ k ][ i ] ;
            }
	  }

          for ( k = i ; k < _m ; k++ ) {
            _A[ k ][ i ] *= scale ;
          }
        }
      }

      _W[ i ] = scale * g ;

      g = s = scale = 0.0 ;

      if ( ( ( i + 1 ) <= _m ) && ( ( i + 1 ) != _n ) ) {
        for ( k = l - 1 ; k < _n ; k++ ) {
          scale += fabs( _A[ i ][ k ] ) ;
        }
        if ( scale != 0.0 ) {
          for ( k = l - 1 ; k < _n ; k++ ) {
            _A[ i ][ k ] /= scale ;
            s += _A[ i ][ k ] * _A[ i ][ k ] ;
          }

          f = _A[ i ][ l - 1 ] ;
          g = -sign( sqrt( s ) , f ) ;
          h = f * g - s ;
          _A[ i ][ l - 1 ] = f - g ;

          for ( k = l - 1 ; k < _n ; k++ ) {
            rv1[ k ] = _A[ i ][ k ] / h ;
          }

	  for ( j = l - 1 ; j < _m ; j++ ) {
	    for ( s = 0.0 , k = l - 1 ;  k < _n ; k++ ) {
              s += _A[ j ][ k ] * _A[ i ][ k ] ;
            }
	    for ( k = l - 1 ; k < _n ; k++ ) {
              _A[ j ][ k ] += s * rv1[ k ] ;
            }
	  }

          for ( k = l - 1 ; k < _n ; k++ ) {
            _A[ i ][ k ] *= scale ;
          }
        }
      }
      anorm = MAX( anorm , ( fabs( _W[ i ] ) + fabs( rv1[ i ] ) ) ) ;
    }

    for ( i = _n - 1 ; i >= 0 ; i-- ) {
      if ( i < ( _n - 1 ) ) {
        if ( g != 0.0 ) {
          for ( j = l ; j < _n ; j++ ) {
            _V[ j ][ i ] = ( _A[ i ][ j ] / _A[ i ][ l ] ) / g ;
          }

          for ( j = l ; j < _n ; j++ ) {
            for ( s = 0.0 , k = l ; k < _n ; k++ ) {
              s += _A[ i ][ k ] * _V[ k ][ j ] ;
            }
            for ( k = l ; k < _n ; k++ ) {
              _V[ k ][ j ] += s * _V[ k ][ i ] ;
            }
          }
        }

        for ( j = l ; j < _n ; j++ ) {
          _V[ i ][ j ] = _V[ j ][ i ] = 0.0 ;
        }
      }

      _V[ i ][ i ] = 1.0 ;
      g = rv1[ i ] ;
      l = i ;
    }

    for ( i = MIN( _m , _n ) - 1 ; i >= 0 ; i-- ) {
      l = i + 1 ;
      g = _W[ i ] ;
      for ( j = l ; j < _n ; j++ ) {
        _A[ i ][ j ] = 0.0 ;
      }

      if ( g != 0.0 ) {
        g = 1.0 / g ;
        for ( j = l ; j < _n ; j++ ) {
          for ( s = 0.0 , k = l ; k < _m ; k++ ) {
            s += _A[ k ][ i ] * _A[ k ][ j ] ;
          }

          f = ( s / _A[ i ][ i ] ) * g ;

          for ( k = i ; k < _m ; k++ ) {
            _A[ k ][ j ] += f * _A[ k ][ i ] ;
          }
	}

        for ( j = i ; j < _m ; j++ ) {
          _A[ j ][ i ] *= g ;
        }
      } 
      else {
        for ( j = i ; j < _m ; j++ ) {
          _A[ j ][ i ] = 0.0 ;
        }
      }

      ++_A[ i ][ i ] ;
    }

    for ( k = _n - 1 ; k >= 0 ; k-- ) {
      for ( int its = 0 ; its < 30 ; its++ ) {
        flag = true ;
        for ( l = k ; l >= 0 ; l-- ) {
          nm = l - 1 ;
          if ( ( fabs( rv1[ l ] ) + anorm ) == anorm ) {
            flag = false ;
            break ;
          }

          if ( ( fabs( _W[ nm ] ) + anorm ) == anorm ) {
            break ;
          }
	}

        if ( flag ) {
          c = 0.0 ;
          s = 1.0 ;
          for ( i = l ; i < k + 1 ; i++ ) {
            f = s * rv1[ i ] ;
            rv1[ i ] = c * rv1[ i ] ;
            if ( ( fabs( f ) + anorm ) == anorm ) {
              break ;
            }
            g = _W[ i ] ;
            h = pythag( f , g ) ;
            _W[ i ] = h ;
            h = 1.0 / h ;
            c = g * h ;
            s = -f * h ;
            for ( j = 0 ; j < _m ; j++ ) {
              y = _A[ j ][ nm ] ;
              z = _A[ j ][  i ] ;
              _A[ j ][ nm ] = y * c + z * s ;
              _A[ j ][  i ] = z * c - y * s ;
            }
          }
        }

        z = _W[ k ] ;

        if ( l == k ) {
          if ( z < 0.0 ) {
            _W[ k ] = -z ;
            for ( j = 0 ; j < _n ; j++ ) {
              _V[ j ][ k ] = -_V[ j ][ k ] ;
            }
          }
          break ;
        }

        if ( its == 29 ) {
          std::stringstream ss(
			       std::stringstream::in | 
			       std::stringstream::out
			      ) ;
	  ss << "No convergence in 30 iterations" ;
	  throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
        }

        x = _W[ l ] ;
        nm = k - 1 ;
        y = _W[ nm ] ;
        g = rv1[ nm ] ;
        h = rv1[ k ] ;
        f = ( ( y - z ) * ( y + z ) + ( g - h ) * ( g + h ) ) / ( 2.0 * h * y ) ;
        g = pythag( f , 1.0 ) ;
        f = ( ( x - z ) * ( x + z ) + h * ( ( y / ( f + sign( g , f ) ) ) - h ) ) / x ;
        c = s = 1.0 ;
        for ( j = l ; j <= nm; j++ ) {
          i = j + 1 ;
          g = rv1[ i ] ;
          y = _W[ i ] ;
          h = s * g ;
          g = c * g ;
          z = pythag( f , h ) ;
          rv1[ j ] = z ;
          c = f / z ;
          s = h / z ;
          f = x * c + g * s ;
          g = g * c - x * s ;
          h = y * s ;
          y *= c ;
          for ( int jj = 0 ; jj < _n ; jj++ ) {
            x = _V[ jj ][ j ] ;
            z = _V[ jj ][ i ] ;
            _V[ jj ][ j ] = x * c + z * s ;
            _V[ jj ][ i ] = z * c - x * s ;
          }

          z = pythag( f , h) ;

          _W[ j ] = z ;
          if ( z != 0.0 ) {
            z = 1.0 / z ;
            c = f * z ; 
            s = h * z ;
	  }

          f = c *  g + s * y ;
          x = c * y - s * g ;
          for ( int jj = 0 ; jj < _m ; jj++ ) {
            y = _A[ jj ][ j ] ;
            z = _A[ jj ][ i ] ;
            _A[ jj ][ j ] = y * c + z * s ;
            _A[ jj ][ i ] = z * c - y * s ;
          }
        }

        rv1[ l ] = 0.0 ;
        rv1[ k ] = f ;
        _W[ k ] = x ;
      }
    }

    // release memory allocated for "rv1"
    delete[] rv1 ;
  }


  /** 
   * \fn void SVDdcmp::bksb( double* b , double* x ) const
   *
   * \brief Computes the solution of the linear system \f$ A \cdot X
   * =  B \f$  by backsubstitution  using the  SVD  decomposition of
   * matrix \f$ A \f$.
   *
   * \param b The vector \f$ B \f$ of the linear system \f$ A \cdot X = B \f$.
   * \param x The solution vector of the linear system \f$ A \cdot X = B \f$.
   */
  void
  SVDdcmp::bksb(
		double* b,
		double* x
               )
    const
  {
    double* tmp = new double[ _n ] ;
    for ( int j = 0 ; j < _n ; j++ ) {
      double s = 0.0 ;
      if ( _W[ j ] != 0.0 ) {
        for ( int i = 0 ; i < _m ; i++ ) {
          s += _A[ i ][ j ] * b[ i ] ;
        }

        s /= _W[ j ] ;
      }

      tmp[ j ] = s ;
    }

    for ( int j = 0 ; j < _n ; j++ ) {
      double s = 0.0 ;
      for ( int jj = 0 ; jj < _n ; jj++ ) {
        s += _V[ j ][ jj ] * tmp[ jj ] ;
      }
      x[ j ] = s ;
    }

    // release memory allocated for "tmp"
    delete[] tmp;
  }


  /** 
   * \fn double SVDdcmp::pythag( double a , double b ) const
   *
   * \brief An  auxiliary procedure for robustly  computing the square
   * root of  the sum  of the  squares, \f$ \sqrt{a^b  + b^2}  \f$, of
   * reals \f$a\f$ and \f$b\f$.
   *
   * \param  a A  given real  number.
   * \param  b Another  given real number.
   *
   * \return The value of \f$ \sqrt{a^b + b^2} \f$.
   */
  double
  SVDdcmp::pythag(
		  double a,
		  double b
		 )
    const
  {
    double absa , absb ;

    absa = fabs( a ) ;
    absb = fabs( b ) ;
    if (absa > absb) {
      return absa * sqrt( 1.0 + SQR( absb / absa ) ) ;
    }

    return ( absb == 0.0 ) ? 0.0 : ( absb * sqrt( 1.0 + SQR( absa / absb ) ) ) ;
  }


  /** 
   * \fn int SVDdcmp::rank() const
   *
   * \brief  Returns  the rank  of  the  matrix  for which  the  SVD
   * decomposition  has been  computed. The  rank value  returned by
   * this method is  equal to the number of  nonzero singular values.
   *
   * \return The rank of the  matrix for which the SVD decomposition
   * has been computed.
   */
  int
  SVDdcmp::rank() const
  {
    int rank = 0 ;
    for ( int i = 0 ; i < _n ; i++ ) {
      if ( _W[ i ] != 0.0 ) {
	++rank ;
      }
      else {
	break ;
      }
    }

    return rank ;
  }

}         // end of namespace numerics

/** @} */ //end of group class.
