/** 
 * \file svddcmp.hpp
 *  
 * \brief This file contains the definition of a class that contains a
 * method for  computing the Singular  Value Decomposition (SVD)  of a
 * real matrix.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef SVDDCMP_HPP
#define SVDDCMP_HPP

#include "exceptionobject.hpp" // commonlib::ExceptionObject


/** 
 * \defgroup NumericsNameSpace Namespace numerics
 * @{
 */

/**
 * \namespace numerics
 *
 * \brief  The  namespace  numerics  contains  a set  of  classes  and
 * functions for numerical computing.
 */

namespace numerics {

  using common::ExceptionObject ;


  /**
   * \class SVDdcmp
   *
   * \brief This class contains  a collection of methods for computing
   * the Singular Value  Decomposition (SVD) of a real  matrix and for
   * solving  linear systems  using  the matrices  resulting from  the
   * decomposition.
   */
  class SVDdcmp {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn SVDdcmp( double** a , int m , int n ) 
     *
     * \brief Creates an instance of this class.
     *
     * \param a A pointer to a real matrix.
     * \param m The number of rows of the given matrix.
     * \param n The number of columns of the given matrix.
     */
    SVDdcmp( double** a , int m , int n ) ;


    /** 
     * \fn SVDdcmp( const SVDdcmp& s ) 
     *
     * \brief  Creates an instance  of this  class by  cloning another
     * instance of this class.
     *
     * \param s A given instance of this class
     */
    SVDdcmp( const SVDdcmp& s ) ;


    /** 
     * \fn ~SVDdcmp() 
     *
     * \brief Destroys an instance of this class.
     */
    ~SVDdcmp() ;


    /** 
     * \fn double matU( int i , int j ) const throw ( ExceptionObject )
     *
     * \brief Returns the  element of the i-th row  and j-th column of
     * the orthogonal  matrix \f$U\f$ of the SVD  decomposition, \f$ U
     * \Sigma S V^t \f$, of the given real matrix, \f$A\f$.
     *
     * \param i The row index of an element.
     * \param j The column index of an element.
     */
    double matU( int i , int j ) const throw ( ExceptionObject ) ;


    /** 
     * \fn double matV( int i , int j ) const throw ( ExceptionObject )
     *
     * \brief Returns the  element of the i-th row  and j-th column of
     * the orthogonal  matrix \f$V\f$ of the SVD  decomposition, \f$ U
     * \Sigma S V^t \f$, of the given real matrix, \f$A\f$.
     *
     * \param i The row index of an element.
     * \param j The column index of an element.
     */
    double matV( int i , int j ) const throw ( ExceptionObject ) ;


    /** 
     * \fn double vecW( int i ) const throw ( ExceptionObject )
     *
     * \brief Returns the i-th singular value of the given real matrix
     * for which the SVD decomposition has been computed.
     *
     * \param i The index of a singular value.
     */
    double vecW( int i ) const throw ( ExceptionObject ) ;


    /** 
     * \fn void lsq_solve( double* b , double* x ) const
     *
     * \brief Solves the linear least  squares problem \f$\| A \cdot X
     * - B \|  \f$, where \f$ A \f$  is the real matrix  for which the
     * SVD decomposition has  been computed, and \f$ B  \f$ is a given
     * real vector.
     *
     * \param b A pointer to the column vector \f$ B \f$.
     * \param x A pointer to the solution vector  \f$ X \f$.
     */
    void lsq_solve( double* b , double* x ) const ;


    /** 
     * \fn int rank() const
     *
     * \brief  Returns  the rank  of  the  matrix  for which  the  SVD
     * decomposition  has been  computed. The  rank value  returned by
     * this method is  equal to the number of  nonzero singular values.
     *
     * \return The rank of the  matrix for which the SVD decomposition
     * has been computed.
     */
    int rank() const ;


  private:
    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn SVDdcmp() 
     *
     * \brief Creates an instance of this class.
     */
    SVDdcmp() {}


    /** 
     * \fn double matA( int i , int j ) const throw ( ExceptionObject )
     *
     * \brief Returns the  element of the i-th row  and j-th column of
     * the real matrix \f$A\f$ for which the SVD decomposition will be
     * computed.
     *
     * \param i The row index of an element.
     * \param j The column index of an element.
     */
    double matA( int i , int j ) const throw ( ExceptionObject ) ;


    /** 
     * \fn void allocate()
     *
     * \brief Allocates  memory for the matrices  \f$ A \f$  and \f$ V
     * \f$, and the vector \f$ W  \f$, where \f$ V\f$ and \f$ W\f$ are
     * the  orthogonal  matrix  and  the  singular value  of  the  SVD
     * decomposition,  \f$ U  \Sigma  S  V^t \f$,  of  the given  real
     * matrix, \f$A\f$.
     */
    void allocate() ;


    /** 
     * \fn void deallocate()
     *
     * \brief Deallocates the memory  allocated for the matrices \f$ A
     * \f$ and \f$ V \f$, and the vector \f$ W \f$, where \f$ V\f$ and
     * \f$ W\f$  are the orthogonal  matrix and the singular  value of
     * the SVD  decomposition, \f$  U \Sigma S  V^t \f$, of  the given
     * real matrix, \f$A\f$.
     */
    void deallocate() ;


    /** 
     * \fn void decomp() const throw ( ExceptionObject )
     *
     * \brief  Computes  the  SVD  decomposition of  the  given,  real
     * matrix.
     */
    void decomp() const throw ( ExceptionObject ) ;


    /** 
     * \fn void bksb( double* b , double* x ) const
     *
     * \brief Computes the solution of the linear system \f$ A \cdot X
     * =  B \f$  by backsubstitution  using the  SVD  decomposition of
     * matrix \f$ A \f$.
     *
     *
     * \param b The vector \f$ B \f$ of the linear system \f$ A \cdot X = B \f$.
     * \param x The solution vector of the linear system \f$ A \cdot X = B \f$.
     */
    void bksb( double* b , double* x ) const ;


    /** 
     * \fn double pythag( double a , double b ) const
     *
     * \brief An auxiliary procedure for robustly computing the square
     * root of  the sum of the  squares, \f$ \sqrt{a^b +  b^2} \f$, of
     * reals \f$a\f$ and \f$b\f$.
     *
     * \param  a A  given real  number.
     * \param  b Another  given real number.
     *
     * \return The value of \f$ \sqrt{a^b + b^2} \f$.
     */
    double pythag( double a , double b ) const ;


    /** 
     * \fn double sign( double a , double b ) const
     *
     * \brief Computes the signed magnitude of a given number based on
     * the sign of another given number.
     *
     * \param  a A  given real  number.
     * \param  b Another  given real number.
     *
     * \return The signed magnitude of a given number.
     */
    double sign( double a , double b ) const
    {
      return ( b >= 0 ) ? ( ( a >= 0 ) ? a : -a ) : ( ( a >= 0 ) ? -a : a ) ;
    }


    // ---------------------------------------------------------------
    //
    // Private attributes
    //
    // ---------------------------------------------------------------

    double** _A ;   ///< A real matrix for which the SVD decomposition is computed.
    double** _V ;   ///< The orthogonal matrix \f$V\f$ of the SVD, \f$ U \Sigma V^t \f$, of \f$A\f$.
    double*  _W ;   ///< A vector for storing the singular values.
    int _m ;   ///< Number of rows of matrix \f$ A \f$.
    int _n ;   ///< Number of columns of matrix \f$ A \f$.

  } ;


}        // end of namespace numerics


/** @} */ //end of group class.

#endif   // SVDDCMP_HPP
