/** 
 * \file tst-svd.cpp
 *  
 * \brief A driver to test the implementation of the SVD library.
 *
 * \author
 * Marcelo Ferreira Siqueira
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * mfsiqueira at gmail (dot) com
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>           // std::cout, std::endl, std::cerr
#include <ios>                // std::fixed
#include <iomanip>            // std::setprecision()
#include <cstdlib>            // EXIT_SUCCESS, EXIT_FAILURE

#include "exceptionobject.hpp" // common::ExceptionObject
#include "svddcmp.hpp"         // numerics::SVDdcmp


using common::ExceptionObject ;
using numerics::SVDdcmp ;


/**
 * \fn int main()
 *
 * \brief A simple test for an implementation of the SVD library.
 *
 * \return An integer number.
 */
int main() 
{
  //
  // A 2x3 matrix of rank 2
  //

  std::cout << std::endl ;

  std::cout << "SVD decomposition of a 2x3 matrix of rank 2" << std::endl ;

  std::cout << std::endl ;

  int r1 = 2 , c1 = 3 ;
  double** mat1 = new double*[ r1 ] ;
  for ( int i = 0 ; i < r1 ; i++ ) {
    mat1[ i ] = new double[ c1 ] ;
  }

  mat1[ 0 ][ 0 ] = -1 ;
  mat1[ 1 ][ 0 ] =  0 ;
  mat1[ 0 ][ 1 ] =  1 ;
  mat1[ 1 ][ 1 ] = -1 ;
  mat1[ 0 ][ 2 ] =  0 ;
  mat1[ 1 ][ 2 ] =  1 ;

  try 
  {
    std::cout << "Matrix A: " << std::endl ;

    std::cout << std::fixed << std::setprecision( 10 ) ;

    for ( int i = 0 ; i < r1 ; i++ ) {
      for ( int j = 0 ; j < c1 ; j++ ) {
	std::cout << "A[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << mat1[ i ][ j ]
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    SVDdcmp s( mat1 , r1 , c1 ) ;

    std::cout << "Matrix U: " << std::endl ;

    for ( int i = 0 ; i < r1 ; i++ ) {
      for ( int j = 0 ; j < c1 ; j++ ) {
	std::cout << "U[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matU( i , j )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Matrix V^t: " << std::endl ;

    for ( int i = 0 ; i < c1 ; i++ ) {
      for ( int j = 0 ; j < c1 ; j++ ) {
	std::cout << "V^t[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matV( j , i )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Singular values: " << std::endl ;

    for ( int i = 0 ; i < c1 ; i++ ) {
      std::cout << "W[ "
	        << i 
	        << " ] = "
	        << s.vecW( i )
		<< std::endl ; 
    }

    std::cout << std::endl ;

    std::cout << "Rank: " 
	      << s.rank() 
              << std::endl ;

    std::cout << std::endl ;

    for ( int i = 0 ; i < r1 ; i++ ) {
      delete[] mat1[ i ] ;
    }
    
    delete[] mat1 ; 
  }
  catch ( const ExceptionObject& xpt ) {
    for ( int i = 0 ; i < r1 ; i++ ) {
      delete[] mat1[ i ] ;
    }
  
    delete[] mat1 ;

    treat_exception( xpt ) ;
    return EXIT_FAILURE ;
  }

  std::cout << std::endl ;

  //
  // A 3x1 matrix of rank 1
  //

  std::cout << "SVD decomposition of a 3x1 matrix of rank 1" << std::endl ;

  std::cout << std::endl ;

  int r2 = 3 , c2 = 1 ;
  double** mat2 = new double*[ r2 ] ;
  for ( int i = 0 ; i < r2 ; i++ ) {
    mat2[ i ] = new double[ c2 ] ;
  }

  mat2[ 0 ][ 0 ] = -1 ;
  mat2[ 1 ][ 0 ] =  2 ;
  mat2[ 2 ][ 0 ] =  2 ;

  try 
  {
    std::cout << "Matrix A: " << std::endl ;

    std::cout << std::fixed << std::setprecision( 10 ) ;

    for ( int i = 0 ; i < r2 ; i++ ) {
      for ( int j = 0 ; j < c2 ; j++ ) {
	std::cout << "A[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << mat2[ i ][ j ]
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }

    std::cout << std::endl ;

    SVDdcmp s( mat2 , r2 , c2 ) ;

    std::cout << "Matrix U: " << std::endl ;

    for ( int i = 0 ; i < r2 ; i++ ) {
      for ( int j = 0 ; j < c2 ; j++ ) {
	std::cout << "U[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matU( i , j )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Matrix V^t: " << std::endl ;

    for ( int i = 0 ; i < c2 ; i++ ) {
      for ( int j = 0 ; j < c2 ; j++ ) {
	std::cout << "V^t[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matV( j , i )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Singular values: " << std::endl ;

    for ( int i = 0 ; i < c2 ; i++ ) {
      std::cout << "W[ "
	        << i 
	        << " ] = "
	        << s.vecW( i )
		<< std::endl ; 
    }

    std::cout << std::endl ;

    std::cout << "Rank: " 
	      << s.rank() 
              << std::endl ;

    std::cout << std::endl ;

    for ( int i = 0 ; i < r2 ; i++ ) {
      delete[] mat2[ i ] ;
    }
    
    delete[] mat2 ; 
  }
  catch ( const ExceptionObject& xpt ) {
    for ( int i = 0 ; i < r2 ; i++ ) {
      delete[] mat2[ i ] ;
    }
  
    delete[] mat2 ;

    treat_exception( xpt ) ;
    return EXIT_FAILURE ;
  }

  std::cout << std::endl ;

  //
  // A 2x2 matrix of rank 2
  //

  std::cout << "SVD decomposition of a 2x2 matrix of rank 2" << std::endl ;

  std::cout << std::endl ;

  int r3 = 2 , c3 = 2 ;
  double** mat3 = new double*[ r3 ] ;
  for ( int i = 0 ; i < r3 ; i++ ) {
    mat3[ i ] = new double[ c3 ] ;
  }

  mat3[ 0 ][ 0 ] =  1 ;
  mat3[ 1 ][ 0 ] = -2 ;
  mat3[ 0 ][ 1 ] =  3 ;
  mat3[ 1 ][ 1 ] = -1 ;

  try 
  {
    std::cout << "Matrix A: " << std::endl ;

    std::cout << std::fixed << std::setprecision( 10 ) ;

    for ( int i = 0 ; i < r3 ; i++ ) {
      for ( int j = 0 ; j < c3 ; j++ ) {
	std::cout << "A[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << mat3[ i ][ j ]
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }

    SVDdcmp s( mat3 , r3 , c3 ) ;

    std::cout << "Matrix U: " << std::endl ;

    for ( int i = 0 ; i < r3 ; i++ ) {
      for ( int j = 0 ; j < c3 ; j++ ) {
	std::cout << "U[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matU( i , j )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Matrix V^t: " << std::endl ;

    for ( int i = 0 ; i < c3 ; i++ ) {
      for ( int j = 0 ; j < c3 ; j++ ) {
	std::cout << "V^t[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matV( j , i )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Singular values: " << std::endl ;

    for ( int i = 0 ; i < c3 ; i++ ) {
      std::cout << "W[ "
	        << i 
	        << " ] = "
	        << s.vecW( i )
		<< std::endl ; 
    }

    std::cout << std::endl ;

    std::cout << "Rank: " 
	      << s.rank() 
              << std::endl ;

    std::cout << std::endl ;

    for ( int i = 0 ; i < r3 ; i++ ) {
      delete[] mat3[ i ] ;
    }
    
    delete[] mat3 ; 
  }
  catch ( const ExceptionObject& xpt ) {
    for ( int i = 0 ; i < r3 ; i++ ) {
      delete[] mat3[ i ] ;
    }
  
    delete[] mat3 ;

    treat_exception( xpt ) ;
    return EXIT_FAILURE ;
  }

  //
  // Solving a linear system using SVD decompisition.
  //

  std::cout << "Solving a linear system using SVD decomposition" << std::endl ;

  std::cout << std::endl ;

  int r4 = 3 , c4 = 3 ;
  double** mat4 = new double*[ r4 ] ;
  for ( int i = 0 ; i < r4 ; i++ ) {
    mat4[ i ] = new double[ c4 ] ;
  }

  mat4[ 0 ][ 0 ] =  2 ;
  mat4[ 0 ][ 1 ] =  1 ;
  mat4[ 0 ][ 2 ] =  1 ;

  mat4[ 1 ][ 0 ] =  4 ;
  mat4[ 1 ][ 1 ] = -6 ;
  mat4[ 1 ][ 2 ] =  0 ;

  mat4[ 2 ][ 0 ] = -2 ;
  mat4[ 2 ][ 1 ] =  7 ;
  mat4[ 2 ][ 2 ] =  2 ;

  double* vecB = new double[ r4 ] ;

  vecB[ 0 ] =  5 ;
  vecB[ 1 ] = -2 ;
  vecB[ 2 ] =  9 ;

  try 
  {
    std::cout << "Matrix A: " << std::endl ;

    std::cout << std::fixed << std::setprecision( 10 ) ;

    for ( int i = 0 ; i < r4 ; i++ ) {
      for ( int j = 0 ; j < c4 ; j++ ) {
	std::cout << "A[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << mat4[ i ][ j ]
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }

    std::cout << std::endl ;

    SVDdcmp s( mat4 , r4 , c4 ) ;

    std::cout << "Matrix U: " << std::endl ;

    for ( int i = 0 ; i < r4 ; i++ ) {
      for ( int j = 0 ; j < c4 ; j++ ) {
	std::cout << "U[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matU( i , j )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Matrix V^t: " << std::endl ;

    for ( int i = 0 ; i < c4 ; i++ ) {
      for ( int j = 0 ; j < c4 ; j++ ) {
	std::cout << "V^t[ "
		  << i 
		  << " , "
		  << j
		  << " ] = "
		  << s.matV( j , i )
		  << '\t' ; 
      }

      std::cout << std::endl ;
    }  

    std::cout << std::endl ;

    std::cout << "Singular values: " << std::endl ;

    for ( int i = 0 ; i < c4 ; i++ ) {
      std::cout << "W[ "
	        << i 
	        << " ] = "
	        << s.vecW( i )
		<< std::endl ; 
    }

    std::cout << std::endl ;

    std::cout << "Rank: " 
	      << s.rank() 
              << std::endl ;

    std::cout << std::endl ;

    std::cout << "Solving the linear system || AX - B ||: " << std::endl ;

    std::cout << std::endl ;

    std::cout << "Vector B: " << std::endl ;

    for ( int i = 0 ; i < r4 ; i++ ) {
      std::cout << "B[ "
		<< i 
		<< " ] = "
		<< vecB[ i ] 
		<< std::endl ; 
    }

    std::cout << std::endl ;

    double* vecX = new double[ r4 ] ;

    s.lsq_solve( vecB , vecX ) ;

    std::cout << "Solution vector: " << std::endl ;

    for ( int i = 0 ; i < c4 ; i++ ) {
      std::cout << "X[ "
	        << i 
	        << " ] = "
	        << vecX[ i ]
		<< std::endl ; 
    }
    
    std::cout << std::endl ;

    for ( int i = 0 ; i < r4 ; i++ ) {
      delete[] mat4[ i ] ;
    }
    
    delete[] mat4 ;

    delete[] vecB ;

    delete[] vecX ;
  }
  catch ( const ExceptionObject& xpt ) {
    for ( int i = 0 ; i < r3 ; i++ ) {
      delete[] mat4[ i ] ;
    }
  
    delete[] mat4 ;

    delete[] vecB ;

    treat_exception( xpt ) ;
    return EXIT_FAILURE ;
  }

  std::cout << std::endl ;

  return EXIT_SUCCESS ;
}



